﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Globalization;

namespace Jegbor
{
    public enum CellState { BeforeEval, DuringEval, AfterEval };
    public enum EvalType { Addition,Multiplication,Division,Default };
    public class BaseCell
    {
        protected int fDecimals = 2;
        protected CellState cellState = CellState.BeforeEval;
        protected Jegbor parent = null;
        public virtual BaseCell CloneCell()
        {
            BaseCell result = new BaseCell();
            result.cellState = CellState.BeforeEval;
            result.fDecimals = fDecimals;
            result.parent = parent;
            return result;
        }
        public BaseCell()
        {
            parent = null;
            fDecimals = 2;
            cellState = CellState.BeforeEval;
        }
        public BaseCell(Jegbor iparent)
        {
            parent = iparent;
            fDecimals = 2;
            cellState = CellState.BeforeEval;
        }
        public virtual string GetText()
        {
            return "";
        }
        /// <summary>
        /// Beállítja, hogy hány tizedesjegy pontossággal jelenjen meg az eredmény
        /// </summary>
        /// <param name="i"></param>
        public int Decimals
        {
            set
            {
                if ((value < 0) || (value > 3)) throw new ArgumentOutOfRangeException("A tizedesjegyek számának 0 és 3 között kell lennie");
                fDecimals = value;
            }
            get
            {
                return fDecimals;
            }
        }
        public virtual string GetFormatted()
        {
            return "";
        }
        public virtual double Evaluate(EvalType etyp)
        {
            cellState = CellState.AfterEval;
            if (etyp == EvalType.Addition) return 0d;
            if (etyp == EvalType.Multiplication) return 1d;
            if (etyp == EvalType.Default) return 0d;
            throw new ArgumentException("Osztás üres cellával");
        }
        public void ResetEval()
        {
            cellState = CellState.BeforeEval;
        }

    }
    public class NumberCell : BaseCell
    {
        public double value;
        public NumberCell(Jegbor iparent,double init)
        {
            parent = iparent;
            fDecimals = 2;
            cellState = CellState.BeforeEval;
            value=init;
        }
        public override BaseCell CloneCell()
        {
            NumberCell result = new NumberCell(parent,value);
            result.fDecimals = fDecimals;
            return result;
        }

        public override string GetText()
        {
            return value.ToString();
        }
        public override string GetFormatted()
        {
            return Math.Round(Evaluate(EvalType.Default),fDecimals).ToString();
        }
        public override double Evaluate(EvalType etyp)
        {
            cellState = CellState.AfterEval;
            return value;
        }
    }
    public class ExprCell : BaseCell
    {
        public string expr;
        public double value;
        public ExprCell(Jegbor iparent,string init)
        {
            parent = iparent;
            fDecimals = 2;
            cellState = CellState.BeforeEval;
            expr=init;
        }
        public override BaseCell CloneCell()
        {
            ExprCell result = new ExprCell(parent, expr);
            result.fDecimals = fDecimals;
            return result;
        }
        public override string GetText()
        {
            return expr;
        }
        public override double Evaluate(EvalType etyp)
        {
            if (cellState == CellState.DuringEval) throw new ArgumentException("Körkörös hivatkozás");
            if (cellState == CellState.AfterEval) return value;

            cellState = CellState.DuringEval;

            value = parent.ParseFunction(expr.Remove(0,1),EvalType.Default);
            cellState = CellState.AfterEval;
            return value;
        }
        public override string GetFormatted()
        {
            if (cellState == CellState.BeforeEval) Evaluate(EvalType.Default);
            if (cellState == CellState.DuringEval) throw new ArgumentException();
            cellState = CellState.AfterEval;
            return Jegbor.RoundToString(value,fDecimals);
        }
    }
    public partial class Jegbor
    {
        private int fCurrentX = 0;
        private int fCurrentY = 0;
        private BaseCell[,] cells;
        public BaseCell CurrentCell
        {
            get
            {
                return cells[fCurrentX,fCurrentY];
            }
        }
        public void EvaluateCells()
        {
            for (int i = 0; i < 5; i++)
                for (int j = 0; j < 8; j++)
                    cells[i, j].ResetEval();
            for (int i = 0; i < 5; i++)
                for (int j = 0; j < 8; j++)
                    cells[i, j].Evaluate(EvalType.Default);
        }
        public static string RoundToString(double d, int digits)
        {
            return Math.Round(d, digits).ToString();
        }
        public void ResetEval()
        {
            for (int i = 0; i < 5; i++)
                for (int j = 0; j < 8; j++)
                    cells[i, j].ResetEval();
        }
        public int CurrentX
        {
            get{
                return fCurrentX;
            }
            set{
                if ((value<0)||(value>=5)) new ArgumentOutOfRangeException("A mező kordináta érvénytelen");
                fCurrentX = value;
            }
        }
        public int CurrentY
        {
            get{
                return fCurrentY;
            }
            set{
                if ((value<0)||(value>=8)) new ArgumentOutOfRangeException("A mező kordináta érvénytelen");
                fCurrentY = value;
            }
        }
        public BaseCell GetCell(int i, int j)
        {
            if ((i < 0) || (i >= 5) || (j < 0) || (j >= 8)) throw new ArgumentOutOfRangeException("Táblázaton kívüli tartomány");
            return cells[i, j];
        }
        public Jegbor()
        {
            cells = new BaseCell[5,8];
            for (int i=0;i<5;i++)
            for (int j=0;j<8;j++)
                cells[i,j]=new BaseCell(this);
        }
        public void SetCurrentCell(string input)
        {
            SetCell(fCurrentX, fCurrentY, input);
        }
        public void SetCell(int col,int row,string expr)
        {
            BaseCell tmp = cells[col, row];
            BaseCell tmp2 = null;
            if (expr.Length < 1)
            {
                tmp2 = new BaseCell(this);
                tmp2.Decimals = tmp.Decimals;
                cells[col, row] = tmp2;
                return;
            }
            if (expr[0] == '=')
            {
                tmp2 = new ExprCell(this,expr);
                tmp2.Decimals = tmp.Decimals;
                cells[col, row] = tmp2;
                return;
            }
            double ival = 0;
            expr = expr.Replace(',','.');
            if (!double.TryParse(expr, out ival))
                throw new ArgumentException("Hibás szám formátum");
            tmp2 = new NumberCell(this, ival);
            tmp2.Decimals = tmp.Decimals;
            cells[col, row] = tmp2;
        }

        public static void ExtractCellPos(string id, ref int col, ref int row)
        {
            if (id.Length != 2) throw new ArgumentException("Érvénytelen mező azonosító \"" + id + "\"");
            int i = 0, j = 0;
            switch (id[0])
            {
                case 'A': i = 0; break;
                case 'B': i = 1; break;
                case 'C': i = 2; break;
                case 'D': i = 3; break;
                case 'E': i = 4; break;
                default:
                    throw new ArgumentException("Érvénytelen mező azonosító \"" + id + "\"");
            }
            switch (id[1])
            {
                case '1': j = 0; break;
                case '2': j = 1; break;
                case '3': j = 2; break;
                case '4': j = 3; break;
                case '5': j = 4; break;
                case '6': j = 5; break;
                case '7': j = 6; break;
                case '8': j = 7; break;
                default:
                    throw new ArgumentException("Érvénytelen mező azonosító \"" + id + "\"");
            }
            col = i;
            row = j;
        }
        public void SetCurrentCellPosition(string id)
        {
            int i = 0, j = 0;
            ExtractCellPos(id, ref i, ref j);
            fCurrentX = i;
            fCurrentY = j;
        }
        public void ExecuteCommand(string command)
        {
            string basecommand = command;
            command=command.ToUpper();
            if (command.Length > 0)
            {
                switch (command[0])
                {
                    case 'G':
                        SetCurrentCellPosition(command.Remove(0, 2));
                        break;
                    case 'S':
                        Save(command.Remove(0, 2));
                        break;
                    case 'L':
                        Load(command.Remove(0, 2));
                        break;
                    case 'F':
                        int tmp = 0;
                        if (!int.TryParse(command.Remove(0, 2).Replace(',','.'), out tmp))
                            throw new ArgumentException("A parancs hibás: \"" + basecommand + "\"");
                        CurrentCell.Decimals=tmp;
                        break;
                    case 'C':
                        string[] coms = command.Remove(0, 2).Split(';');
                        if (coms.Length != 2) throw new ArgumentException("A paraméternek pontosan egy ;-t kell tartalmaznia");
                        int ax = 0, ay = 0, bx = 0, by = 0;
                        ExtractCellPos(coms[0], ref ax, ref ay);
                        ExtractCellPos(coms[1], ref bx, ref by);
                        cells[bx, by] = cells[ax, ay].CloneCell();
                        break;
                    default:
                        SetCurrentCell(command);
                        break;

                }
            }
        }
        public static string DoubleToString(double d)
        {
            string result = d.ToString();
            return result.Replace(',', '.');
        }
        public BaseCell GetCell(string id)
        {
            int i=0;
            int j=0;
            ExtractCellPos(id, ref i, ref j);
            return cells[i,j];
        }
    }
}